package rione.agent;

import java.util.*;

import rescuecore2.messages.Command;
import rescuecore2.standard.components.*;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.ChangeSet;
import rione.object.AdvancedWorldModel;

public abstract class Agent<E extends StandardEntity> extends StandardAgent<E> {

	/** 標準のタイムステップ数 300[step]*/
	private static final int DEFAULT_TIMESTEP = 300;
	
	/** 標準のエージェント思考可能時間 1000[ms] */
	private static final int DEFAULT_THINK_TIME = 1000;

	/** タイムステップのキー */
	private static final String TIMESTEP_KEY = "kernel.timesteps";
	
	/** エージェント思考可能時間のキー */
	private static final String THINK_TIME_KEY = "kernel.agents.think-time";

	/** WorldModelへの参照 */
	protected AdvancedWorldModel world;

	/** 現在の時間 */
	protected int time;

	protected ChangeSet changed;

	/** 各サイクルに送信したコマンドの履歴 <時間, コマンド>*/
	private Map<Integer, StandardMessageURN> commandHistory = new HashMap<Integer, StandardMessageURN>();

	/** シミュレーションを行うサイクル数  */
	protected int timestep;

	/** エージェントの思考可能時間  */
	protected int thinkTime;

	@Override   
	protected void postConnect() {
		super.postConnect();
		initialize();
	}

	/**
	 * エージェントの初期化を行う。
	 * 必要な時にオーバーライドして使う。
	 * オーバーライドする際は、メソッド内で必ずスーパークラスのinitialize()を呼び出すこと。
	 */
	protected void initialize() {
		timestep = config.getIntValue(TIMESTEP_KEY, DEFAULT_TIMESTEP);
		thinkTime = config.getIntValue(THINK_TIME_KEY, DEFAULT_THINK_TIME);

		model.indexClass(
				StandardEntityURN.AMBULANCE_TEAM,
				StandardEntityURN.FIRE_BRIGADE,
				StandardEntityURN.POLICE_FORCE,
				StandardEntityURN.BUILDING,
				StandardEntityURN.AMBULANCE_CENTRE,
				StandardEntityURN.FIRE_STATION,
				StandardEntityURN.POLICE_OFFICE,
				StandardEntityURN.ROAD,
				StandardEntityURN.REFUGE,
				StandardEntityURN.BUILDING,
				StandardEntityURN.BLOCKADE,
				StandardEntityURN.CIVILIAN );
		world = (AdvancedWorldModel)model;
		world.initialize(timestep, config);
	}

	@Override
	protected StandardWorldModel createWorldModel() {
		return new AdvancedWorldModel();
	}

	protected abstract void act() throws ActionCommandException;

	/**
	 * actの前に呼び出すメソッド。
	 * 必要な時にオーバーライドして使う。
	 * メソッド内でスーパークラスのこのメソッドを呼び出すこと。
	 */
	protected void prepareForAct() {}

	/**
	 * コマンドを送信した後に呼び出すメソッド。
	 * 必要な時にオーバーライドして使う。
	 * メソッド内でスーパークラスのこのメソッドを呼び出すこと。
	 */
	protected void afterAct() {}

	protected abstract void sense(ChangeSet changed);
	protected abstract void hear(Collection<Command> heard);

	@Override
	protected void think(int time, ChangeSet changed, Collection<Command> heard) {
		try { // 試合時にはこのコメントアウトを外す
			this.time = time;
			world.time = time;
			this.changed = changed;
			sense(changed);
			hear(heard);
			prepareForAct();
			try {
				act();
				rest();
			} catch (ActionCommandException ace) {
				commandHistory.put(time, ace.message);
			}
			afterAct();
			world.time++;	// thinkメソッドが呼ばれる前にWorldModelのtimeが参照される可能性があるので、1足しておく
		} catch (Exception e) { // 試合時にはこのコメントアウトを外す
			System.err.println("Exception : time = "+time+" id = "+me().getID());
			e.printStackTrace();
		} catch (Error e) {
			System.err.println("Error : time = "+time+" id = "+me().getID());
			e.printStackTrace();
		}
	}

	protected void rest() throws ActionCommandException {
		sendRest(time);
		throw new ActionCommandException(StandardMessageURN.AK_REST);
	}

	protected StandardMessageURN getCommandHistory(int time) {
		return commandHistory.get(time);
	}

	public int getTimestep() {
		return timestep;
	}

	@SuppressWarnings("serial")
	public
	static class ActionCommandException extends Exception {
		private StandardMessageURN message;
		public ActionCommandException(StandardMessageURN message) { 
			super(); 
			this.message = message;
		}
	}
}
