package rione.agent;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.messages.AKMove;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.object.ReachableArea;

public abstract class HumanoidAgent<E extends StandardEntity> extends CommunicationAgent<E> {

	protected ReachableArea reachable;
	
	@Override
	protected void initialize() {
		super.initialize();
		world.setMe((Human)me());
		world.setIsolationEntities();

        reachable = new ReachableArea(world);
        world.setReachable(reachable);
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();

		updateReachableArea();
	}
	

	/**
	  * パスと目標地点の座標を渡したmove
	  * @param path 目標地点までのパス
	  * @param x 目標地点のX座標
	 * @throws ActionCommandException
	 */
	protected void move(List<EntityID> path, int x, int y) throws ActionCommandException {
		send(new AKMove(getID(), time, path, x, y));
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}
	
	/**
	 * 目的AreaとXY座標を指定したmove
	 * @param destination 目的Area
	 * @param x 目的X座標
	 * @param y 目的Y座標
	 * @throws ActionCommandException
	 * この移動方法を指定したい場合は，目標地点をArea型で指定する．
	 */
	protected void move(Area destination, int x, int y) throws ActionCommandException {
		move(AreaRouter.get((Area)location(), destination, CostFunction.normal(reachable, world), world, thinkTime), x, y);
	}
	
	/**
	 * パスと目標地点の座標を渡したmove
	 * @param path 目標地点までのパス
	 * @param location 目標地点のX,Y座標
	 * @throws ActionCommandException
	 */
	protected void move(List<EntityID> path, Pair<Integer, Integer> location) throws ActionCommandException {
		if (location == null) {
			move(path, -1, -1);
		} else {
			move(path, location.first(), location.second());
		}
	}
	
	/**
	 * パスを渡したmove
	 * @param path 目標地点までのパス
	 * @throws ActionCommandException
	 */
	protected void move(List<EntityID> path) throws ActionCommandException {
		move(path, -1, -1);
	}
	
	/**
	 * 目標エンティティとコストを渡したmove
	 * @param destination
	 * @param costFunc
	 * @throws ActionCommandException
	 */
	protected void move(StandardEntity destination, CostFunction costFunc)  throws ActionCommandException {
		if (destination instanceof Human) {
			Human human = (Human)destination;
			StandardEntity position = human.getPosition(world);
			if (position instanceof Area) {
				move(AreaRouter.get((Area)location(), (Area)position, costFunc, world, thinkTime), destination.getLocation(world));
			} else {
				// なんらかの理由でエージェントがATの上にある場合，
				// もしくは現在位置がnullの場合エラーになる
				if (position == null) {
					System.err.println("Move error\ttime="+time+" id="+me().getID()+" my position is null");
				} else {
					System.err.println("Move error\ttime="+time+" id="+me().getID()+" move from "+position);
				}
				return;
			}
		} else if (destination instanceof Blockade) {
			Blockade block = (Blockade)destination;
			EntityID position = block.getPosition();
			move(AreaRouter.get((Area)location(), (Area)world.getEntity(position), costFunc, world, thinkTime), destination.getLocation(world));
		}
		move(AreaRouter.get((Area)location(), (Area)destination, costFunc, world, thinkTime), destination.getLocation(world));
	}
	
	/**
	 * 目標エンティティを渡したmove
	 * @param destination
	 * @throws ActionCommandException
	 */
	protected void move(StandardEntity destination)  throws ActionCommandException {
		move(destination, CostFunction.normal(reachable, world));
	}
	
	/**
	 * 目標のEntityIDとコストを渡したmove
	 * @param destination
	 * @param costFunc
	 * @throws ActionCommandException
	 */
	protected void move(EntityID destination, CostFunction costFunc) throws ActionCommandException {
		move(world.getEntity(destination), costFunc);
	}
	
	/**
	 * 目標のEntityIDを渡したmove
	 * @param destination
	 * @throws ActionCommandException
	 */
	protected void move(EntityID destination) throws ActionCommandException {
		move(world.getEntity(destination));
	}

	protected void move(Collection<? extends StandardEntity> destinations, CostFunction costFunc) throws ActionCommandException {
		Set<StandardEntity> destinations_set = new HashSet<StandardEntity>(destinations);
		move(AreaRouter.get((Area)location(), destinations_set, costFunc, world, thinkTime));
	}
	
	protected void move(Collection<? extends StandardEntity> destinations) throws ActionCommandException {
		move(destinations, CostFunction.normal(reachable, world));
	}
	
	private void updateReachableArea() {
		for (EntityID id : getVisibleEntitiesID()) {
			StandardEntity entity = world.getEntity(id);
			if (entity instanceof Road) {
				reachable.update((Road)entity);
			}
		}
		reachable.updateReachable();
		//reachable.resetPassable();
	}
}
