package rione.communication;

import rescuecore2.config.Config;
import rescuecore2.registry.Registry;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.properties.EntityRefProperty;
import rescuecore2.worldmodel.properties.IntProperty;

import rione.object.AdvancedWorldModel;

/**
 * 通信情報を読み込み、WorldModelに反映させるクラス
 * MessageWriterで情報を書き込む際は、読み込む順番と同じ順番で書き込まなければならない
 * @author com
 *
 */
public class MessageReader extends MessageConstants {

	/** メッセージの送信者*/
	private EntityID sender;
	/** メッセージ が送られてきたチャンネル*/
	private int channel;
	/** メッセージが送られた時間 */
	private int sendTime;
	/** ビット列のストリーム */
	private BitArrayInputStream in;

	/**
	 * Civilianのものか
	 * @deprecated
	 * 	このクラスではCivilianの情報でないということを保持するために使っています<br>
	 * 	Civilianの情報はこのクラスには渡さないでください
	 */
	@Deprecated
	private boolean isCivilian;
	
	/**
	 * コンストラクタ
	 * @param world
	 */
	public MessageReader(AdvancedWorldModel world, Config config) {
		super(world, config);
	}
	
	/**
	 * 通信情報を読み込む準備をする
	 * @param message
	 */
	public void open(AKSpeak message) {
		this.sender = message.getAgentID();
		this.channel = message.getChannel();
		this.sendTime = message.getTime();
		
		StandardEntity entity = world.getEntity(sender); 
		if (channel == 0 && (entity == null || entity instanceof Civilian)) {
			isCivilian = true;
		} else {
			isCivilian = false;
			this.in = new BitArrayInputStream(message.getContent());
		}
	}
	
	/**
	 * 通信情報を読み込み、WorldModelに反映させる
	 */
	public void read() {
		int header = in.readBit(OBJECT_HEADER_BIT_SIZE);
		switch (header) {
		case HEADER_BUILDING :
			readBuilding();
			break;
		case HEADER_BLOCKADE :
			readBlockade();
			break;
		case HEADER_CIVILIAN :
			readCivilian();
			break;
		case HEADER_PLATOON :
			readPlatoon();
			break;
		case HEADER_ROAD :
			readRoad();
			break;
		default :
			// ignore other
			break;
		}
	}
	
	/**
	 * ストリームに読み込める情報が残っているかどうか
	 * @return
	 */
	public boolean hasNext() {
		return !isCivilian && in.hasNext();
	}
	
	private void readRoad() {
		EntityID id = world.getEntityIDfromUniform(StandardEntityURN.ROAD, in.readBit(ROADS_BIT_SIZE)); //uniform	
    	boolean passable = in.readBit(ROAD_PASSABLE_BIT_SIZE)==1 ? true : false;

    	if (world.getLastSeenTime(id) < sendTime) {
        	Entity existingEntity = getEntity(id, StandardEntityURN.ROAD);
        	world.getReachable().update((Road)existingEntity, passable);
    	}
	}
	
	private void readBuilding() {
		EntityID id = world.getEntityIDfromUniform(StandardEntityURN.BUILDING, in.readBit(BUILDINGS_BIT_SIZE)); //uniform	
    	int fieryness = in.readBit(BUILDING_FIERYNESS_BIT_SIZE);
    	
    	// fierynessの反映
    	try {
    		Entity existingEntity = getEntity(id, StandardEntityURN.BUILDING);
    		int currentFieryness = ((Building) existingEntity).getFieryness();
    		if ((!StandardEntityConstants.Fieryness.values()[currentFieryness].equals(Fieryness.BURNT_OUT))
    				&& (world.getLastSeenTime(id) < sendTime)) {
            	takeIntValue(existingEntity, StandardPropertyURN.FIERYNESS, fieryness);
            	if (0 < fieryness) {
            		world.setBurningBuildings((Building)existingEntity);
            	}
        	}
    	}
    	catch(NullPointerException e) {}
    	catch(ClassCastException e) {}
    	
	}
	
	private void readBlockade() {
		EntityID id = new EntityID(in.readBit(BlOCKADE_ID_BIT_SIZE));		
		int x = decodeX(in.readBit(X_BIT_SIZE));
		int y = decodeY(in.readBit(Y_BIT_SIZE));

		if (world.getLastSeenTime(id) < sendTime) {
			Entity existingEntity = getEntity(id, StandardEntityURN.BLOCKADE);	
			
			// Apexes情報がnullだと、world.getObjectInRangeメソッドを呼び出した際に
			// エラーとなるので、とりあえず適当な値を入れておく
			Blockade block = (Blockade)existingEntity;
			block.getApexesProperty().setValue(new int[]{0});

			// 座標からAreaを得る
			Area location = world.getPositionFromCoordinates(x, y);
			if (location != null) {
				if (location instanceof Road) { // U,L字型の閉塞は、中心点がBuildingである可能性がある
					// AreaのEntityRefListPropertyに閉塞を追加
					location.getBlockadesProperty().addValue(id);
				}
				// WorldModelに情報を反映
				takeEntityRefValue(existingEntity, StandardPropertyURN.POSITION, location.getID());
			}
			takeIntValue(existingEntity, StandardPropertyURN.X, x);
			takeIntValue(existingEntity, StandardPropertyURN.Y, y);
		}
	}

	private void readCivilian() {
		EntityID id = new EntityID(in.readBit(CIVILIAN_ID_BIT_SIZE));
		int hp = decodeHP(in.readBit(HUMAN_HP_BIT_SIZE));
		int damage = decodeDamage(in.readBit(HUMAN_DAMAGE_BIT_SIZE));
		int buriesness = in.readBit(HUMAN_BURIEDNESS_BIT_SIZE);
		int x = decodeX(in.readBit(X_BIT_SIZE));
		int y = decodeY(in.readBit(Y_BIT_SIZE));

		if (world.getLastSeenTime(id) < sendTime) {
			Entity existingEntity = getEntity(id, StandardEntityURN.CIVILIAN);
			takeIntValue(existingEntity, StandardPropertyURN.HP, hp);
			takeIntValue(existingEntity, StandardPropertyURN.DAMAGE, damage);
			takeIntValue(existingEntity, StandardPropertyURN.BURIEDNESS, buriesness);
			takeIntValue(existingEntity, StandardPropertyURN.X, x);
			takeIntValue(existingEntity, StandardPropertyURN.Y, y);
			Area position = world.getPositionFromCoordinates(x, y);
			if (position != null) {
				takeEntityRefValue(existingEntity, StandardPropertyURN.POSITION, position.getID());
			} else {
				existingEntity.getProperty(StandardPropertyURN.POSITION.toString()).undefine();
			}
		}
	}
	
	private void readPlatoon() {
		switch (in.readBit(PLATOON_HEADER_BIT_SIZE)) {
		case PLATOON_TYPE_FB :
			readFireBrigade();
			break;
		case PLATOON_TYPE_PF :
			readPoliceForce();
			break;
		case PLATOON_TYPE_AT :
			readAmbulanceTeam();
			break;
		default :
			break;
		}
	}
	
	public void readFireBrigade() {
		EntityID id = world.getEntityIDfromUniform(StandardEntityURN.FIRE_BRIGADE, in.readBit(FB_UNIFORM_BIT_SIZE));
		int hp = decodeHP(in.readBit(HUMAN_HP_BIT_SIZE));
		int damage = decodeDamage(in.readBit(HUMAN_DAMAGE_BIT_SIZE));
		int buriesness = in.readBit(HUMAN_BURIEDNESS_BIT_SIZE);
		int x = decodeX(in.readBit(X_BIT_SIZE));
		int y = decodeY(in.readBit(Y_BIT_SIZE));
		int water = in.readBit(PROPERTY_FB_WATER_BIT_SIZE)*100;
		boolean isLocateInBlockade = in.readBit(PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE)==1?true:false;

		Entity existingEntity = getEntity(id, StandardEntityURN.FIRE_BRIGADE);
		takeIntValue(existingEntity, StandardPropertyURN.WATER_QUANTITY, water);
		if (world.getLastSeenTime(id) < sendTime) {
			takeIntValue(existingEntity, StandardPropertyURN.HP, hp);
			takeIntValue(existingEntity, StandardPropertyURN.DAMAGE, damage);
			takeIntValue(existingEntity, StandardPropertyURN.BURIEDNESS, buriesness);
			takeIntValue(existingEntity, StandardPropertyURN.X, x);
			takeIntValue(existingEntity, StandardPropertyURN.Y, y);
			Area position = world.getPositionFromCoordinates(x, y);
			if (position != null) {
				takeEntityRefValue(existingEntity, StandardPropertyURN.POSITION, position.getID());
			} else {
				existingEntity.getProperty(StandardPropertyURN.POSITION.toString()).undefine();
			}
		}
		if (isLocateInBlockade) {
			world.addLocateInBlockadeAgents((Human)existingEntity);
		}
		else {
			world.removeLocateInBlockadeAgents((Human)existingEntity);
		}
	}
	
	public void readPoliceForce() {
		EntityID id = world.getEntityIDfromUniform(StandardEntityURN.POLICE_FORCE, in.readBit(PF_UNIFORM_BIT_SIZE));
		int hp = decodeHP(in.readBit(HUMAN_HP_BIT_SIZE));
		int damage = decodeDamage(in.readBit(HUMAN_DAMAGE_BIT_SIZE));
		int buriesness = in.readBit(HUMAN_BURIEDNESS_BIT_SIZE);
		int x = decodeX(in.readBit(X_BIT_SIZE));
		int y = decodeY(in.readBit(Y_BIT_SIZE));

		Entity existingEntity = getEntity(id, StandardEntityURN.POLICE_FORCE);
		if (world.getLastSeenTime(id) < sendTime) {
			takeIntValue(existingEntity, StandardPropertyURN.HP, hp);
			takeIntValue(existingEntity, StandardPropertyURN.DAMAGE, damage);
			takeIntValue(existingEntity, StandardPropertyURN.BURIEDNESS, buriesness);
			takeIntValue(existingEntity, StandardPropertyURN.X, x);
			takeIntValue(existingEntity, StandardPropertyURN.Y, y);
			Area position = world.getPositionFromCoordinates(x, y);
			if (position != null) {
				takeEntityRefValue(existingEntity, StandardPropertyURN.POSITION, position.getID());
			} else {
				existingEntity.getProperty(StandardPropertyURN.POSITION.toString()).undefine();
			}
		}
	}
	
	public void readAmbulanceTeam() {
		EntityID id = world.getEntityIDfromUniform(StandardEntityURN.AMBULANCE_TEAM, in.readBit(AT_UNIFORM_BIT_SIZE));
		int hp = decodeHP(in.readBit(HUMAN_HP_BIT_SIZE));
		int damage = decodeDamage(in.readBit(HUMAN_DAMAGE_BIT_SIZE));
		int buriesness = in.readBit(HUMAN_BURIEDNESS_BIT_SIZE);
		int x = decodeX(in.readBit(X_BIT_SIZE));
		int y = decodeY(in.readBit(Y_BIT_SIZE));
		boolean isLocateInBlockade = in.readBit(PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE)==1?true:false;

		Entity existingEntity = getEntity(id, StandardEntityURN.AMBULANCE_TEAM);
		if (world.getLastSeenTime(id) < sendTime) {
			takeIntValue(existingEntity, StandardPropertyURN.HP, hp);
			takeIntValue(existingEntity, StandardPropertyURN.DAMAGE, damage);
			takeIntValue(existingEntity, StandardPropertyURN.BURIEDNESS, buriesness);
			takeIntValue(existingEntity, StandardPropertyURN.X, x);
			takeIntValue(existingEntity, StandardPropertyURN.Y, y);
			Area position = world.getPositionFromCoordinates(x, y);
			if (position != null) {
				takeEntityRefValue(existingEntity, StandardPropertyURN.POSITION, position.getID());
			} else {
				existingEntity.getProperty(StandardPropertyURN.POSITION.toString()).undefine();
			}
		}
		if (isLocateInBlockade) {
			world.addLocateInBlockadeAgents((Human)existingEntity);
		}
		else {
			world.removeLocateInBlockadeAgents((Human)existingEntity);
		}
	}
	
	private int decodeHP(int hp) {
		return hp * HP_PERSEPTION;
	}
	
	private int decodeDamage(int damage) {
		return damage * DAMAGE_PERSEPTION;
	}
	
	private int decodeX(int x) {
		return x + MIN_X;
	}
	
	private int decodeY(int y) {
		return y + MIN_Y;
	}
		
	/**
	 * EntityにIntProperty型の通信情報を反映させる
	 * @param existingEntity
	 * @param urn
	 * @param value
	 */
	private void takeIntValue(Entity existingEntity, StandardPropertyURN urn, int value) {
		if (world.getPropertyTimeStamp(existingEntity.getID(), urn) < sendTime) { // タイムスタンプチェック
			((IntProperty)existingEntity.getProperty(urn.toString())).setValue(value);
			world.setProtertyTimestamp(existingEntity.getID(), urn, sendTime); // タイムスタンプ更新
		}
	}
	
	/**
	 * EntityにEntityRefProperty型の通信情報を反映させる
	 * @param existingEntity
	 * @param urn
	 * @param value
	 */
	private void takeEntityRefValue(Entity existingEntity, StandardPropertyURN urn, EntityID value) {
		if (world.getPropertyTimeStamp(existingEntity.getID(), urn) < sendTime) {
			((EntityRefProperty)existingEntity.getProperty(urn.toString())).setValue(value);
			world.setProtertyTimestamp(existingEntity.getID(), urn, sendTime);
		}
	}
	
	/**
	 * 未知である可能性のあるEntityを取得する(Blockade or Civilian)
	 * @param id 取得するEntityのID
	 * @param urn 取得するEntityの種類
	 * @return Entity
	 */
	private Entity getEntity(EntityID id, StandardEntityURN urn) {
        Entity existingEntity = world.getEntity(id); // Entityを取得
        if (existingEntity == null) { // 未知のEntityの場合
            // 新しくEntityを作る
            existingEntity = Registry.getCurrentRegistry().createEntity(urn.toString(), id);
            world.addStateChangeListener(existingEntity);
            world.addEntity(existingEntity);
        }
        return existingEntity;
	}
}
