package rione.communication;

import java.util.*;

import rescuecore2.config.Config;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.agent.AgentConstants;
import rione.object.AdvancedWorldModel;

/**
 * Entity等の情報を通信用のビット列に書き込むクラス
 * MessageReaderで情報を読み込む際は、書き込んだ順番と同じ順番で読み込まなければならない
 * @author com
 *
 */
public class MessageWriter extends MessageConstants {
	
	private BitArrayOutputstream out = new BitArrayOutputstream();
	private Set<StandardEntity> wrote = new HashSet<StandardEntity>();
	
	/**
	 * コンストラクタ
	 * @param model
	 */
	public MessageWriter(AdvancedWorldModel model, Config config) {
		super(model, config);
	}
	
	/**
	 * MessageWriterを初期化する
	 * サイクルの頭に毎回呼び出す
	 */
	public void init() {
		out.clear();
		wrote.clear();
	}
	
	/**
	 * ストリームに書き込まれたビット列を返す
	 * @return
	 */
	public byte[] getData() {
		return out.getData();
	}

	/**
	 * ストリームに書き込まれたビット列のバイトサイズを返す
	 * @return
	 */
	public int getDatasize() {
		return out.getDatasize();
	}

	/**
	 * 視界情報から得た情報をビット列に書き込む
	 * @param changeSet
	 */
	public void write(ChangeSet changeSet) {
        for (EntityID id : changeSet.getChangedEntities()) {
        	if (world.time - world.getLastChangedTime(id) < 10) {
        		write(world.getEntity(id));
        	}
        }
	}
	
	public void write(StandardEntity entity) {
		switch (entity.getStandardURN()) {
		case BUILDING :
			writeBuilding((Building)entity);
			break;
		case CIVILIAN :
			writeCivilian((Civilian)entity);
			break;
		case BLOCKADE :
			writeBlockade((Blockade)entity);
			break;
		case ROAD :
			writeRoad((Road)entity);
			break;
		default:
			if (AgentConstants.PLATOONS.contains(entity.getStandardURN())) {
				writePlatoon((Human) entity);
			}
			break;
		}
	}

	private void writeRoad(Road road) {
		out.writeBit(HEADER_ROAD, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(world.getUniform(road.getID()), ROADS_BIT_SIZE); //uniform
		boolean passable = world.getReachable().isPassable(road);
		out.writeBit(passable? 1:0, ROAD_PASSABLE_BIT_SIZE);
	}

	public void writeBuilding(Building building) {
		out.writeBit(HEADER_BUILDING, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(world.getUniform(building.getID()), BUILDINGS_BIT_SIZE); //uniform
		out.writeBit(building.getFieryness(), BUILDING_FIERYNESS_BIT_SIZE); // fieryness
	}
	
	public void writeCivilian(Civilian civilian) {
		if (!civilian.isXDefined() || !civilian.isYDefined()) {
			return;
		}
		out.writeBit(HEADER_CIVILIAN, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(civilian.getID().getValue(), CIVILIAN_ID_BIT_SIZE); // ID
		out.writeBit(encodeHP(civilian.getHP()), HUMAN_HP_BIT_SIZE); //HP
		out.writeBit(encodeDamage(civilian.getDamage()), HUMAN_DAMAGE_BIT_SIZE); /// damage
		out.writeBit(civilian.getBuriedness(), HUMAN_BURIEDNESS_BIT_SIZE); // buriedness
		out.writeBit(encodeX(civilian.getX()), X_BIT_SIZE); // x
		out.writeBit(encodeY(civilian.getY()), Y_BIT_SIZE); // y
	}	
	
	public void writeBlockade(Blockade blockade) {
		out.writeBit(HEADER_BLOCKADE, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(blockade.getID().getValue(), BlOCKADE_ID_BIT_SIZE); // ID
		out.writeBit(encodeX(blockade.getX()), X_BIT_SIZE); // x
		out.writeBit(encodeY(blockade.getY()), Y_BIT_SIZE); // y
	}	
	
	public void writePlatoon(Human platoon) {
		switch (platoon.getStandardURN()) {
		case FIRE_BRIGADE :
			writeFireBrigade((FireBrigade)platoon);
			break;
		case POLICE_FORCE :
			writePoliceForce((PoliceForce)platoon);
			break;
		case AMBULANCE_TEAM :
			writeAmbulanceTeam((AmbulanceTeam)platoon);
		break;
		}
	}
	
	private void writeFireBrigade(FireBrigade fb) {
		out.writeBit(HEADER_PLATOON, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(PLATOON_TYPE_FB, PLATOON_HEADER_BIT_SIZE);
		out.writeBit(world.getUniform(fb.getID()), FB_UNIFORM_BIT_SIZE); // ID
		out.writeBit(encodeHP(fb.getHP()), HUMAN_HP_BIT_SIZE); //HP
		out.writeBit(encodeDamage(fb.getDamage()), HUMAN_DAMAGE_BIT_SIZE); /// damage
		out.writeBit(fb.getBuriedness(), HUMAN_BURIEDNESS_BIT_SIZE); // buriedness
		out.writeBit(encodeX(fb.getX()), X_BIT_SIZE); // x
		out.writeBit(encodeY(fb.getY()), Y_BIT_SIZE); // y
		out.writeBit(fb.getWater()/100, PROPERTY_FB_WATER_BIT_SIZE);
		if (world.isLocateInBlockade(fb)) {
			out.writeBit(1, PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE);
		}
		else {
			out.writeBit(0, PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE);
		}
	}
	
	private void writePoliceForce(PoliceForce pf) {
		out.writeBit(HEADER_PLATOON, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(PLATOON_TYPE_PF, PLATOON_HEADER_BIT_SIZE);
		out.writeBit(world.getUniform(pf.getID()), PF_UNIFORM_BIT_SIZE); // ID
		out.writeBit(encodeHP(pf.getHP()), HUMAN_HP_BIT_SIZE); //HP
		out.writeBit(encodeDamage(pf.getDamage()), HUMAN_DAMAGE_BIT_SIZE); /// damage
		out.writeBit(pf.getBuriedness(), HUMAN_BURIEDNESS_BIT_SIZE); // buriedness
		out.writeBit(encodeX(pf.getX()), X_BIT_SIZE); // x
		out.writeBit(encodeY(pf.getY()), Y_BIT_SIZE); // y
	}
	
	private void writeAmbulanceTeam(AmbulanceTeam at) {
		out.writeBit(HEADER_PLATOON, OBJECT_HEADER_BIT_SIZE);
		out.writeBit(PLATOON_TYPE_AT, PLATOON_HEADER_BIT_SIZE);
		out.writeBit(world.getUniform(at), AT_UNIFORM_BIT_SIZE); // ID
		out.writeBit(encodeHP(at.getHP()), HUMAN_HP_BIT_SIZE); //HP
		out.writeBit(encodeDamage(at.getDamage()), HUMAN_DAMAGE_BIT_SIZE); /// damage
		out.writeBit(at.getBuriedness(), HUMAN_BURIEDNESS_BIT_SIZE); // buriedness
		out.writeBit(encodeX(at.getX()), X_BIT_SIZE); // x
		out.writeBit(encodeY(at.getY()), Y_BIT_SIZE); // y
		if (world.isLocateInBlockade(at)) {
			out.writeBit(1, PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE);
		}
		else {
			out.writeBit(0, PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE);
		}
	}
	
//	public void writeTask(Task task) {
//		out.writeBit(TYPE_TASK, OBJECT_HEADER_SIZE);
//		out.writeBit(task.taskType(), Task.TYPE_BIT_SIZE);
//		out.writeBit(task.entity.getID().getValue(), ENTITY_ID_BIT_SIZE);
//		out.writeBit(task.approach, task.getApproachBitSize());
//		out.writeBit(task.getState(), task.stateBitSize());
//	}
	
	private int encodeHP(int hp) {
		return hp / HP_PERSEPTION;
	}
	
	private int encodeDamage(int damage) {
		return damage < 2000 ? damage /DAMAGE_PERSEPTION : 2000/DAMAGE_PERSEPTION;
	}
	
	private int encodeX(int x) {
		return x - MIN_X;
	}
	
	private int encodeY(int y) {
		return y - MIN_Y;
	}
}
