package rione.communication.channel;

import java.util.*;

import rescuecore2.config.Config;
import rione.agent.CommunicationAgent;
import rione.object.AdvancedWorldModel;

public class ChannelManager {

	// constants //////////////////////////////////////////////////////////////////////////
	private static final String CHANNELS_KEY_PREFIX = "comms.channels.";
	private static final String CHANNELS_TYPE_KEY = ".type";
	private static final String CHANNELS_COUNT_KEY = CHANNELS_KEY_PREFIX + "count";
	private static final String SUBSCRIBE_PLATOON_KEY = CHANNELS_KEY_PREFIX + "max.platoon";
	private static final String SUBSCRIBE_CENTER_KEY = CHANNELS_KEY_PREFIX + "max.centre";

	// voice
	private static final String RANGE_SUFFIX = ".range";
	private static final String MESSAGE_SIZE_SUFFIX = ".messages.size";
	private static final String MESSAGE_MAX_SUFFIX = ".messages.max";

	// radio
	private static final String RADIO_BAND_WIDTH_KEY = ".bandwidth";

	// noise
	private static final String NOISE_INPUT_FAILURE_USE_KEY = ".noise.input.failure.use";
	private static final String NOISE_INPUT_FAILURE_RATIO_KEY = ".noise.input.failure.p";
	private static final String NOISE_INPUT_DROPOUT_USE_KEY = ".noise.input.dropout.use";
	private static final String NOISE_INPUT_DROPOUT_RATIO_KEY = ".noise.input.dropout.p";
	private static final String NOISE_OUTPUT_FAILURE_USE_KEY = ".noise.output.failure.use";
	private static final String NOISE_OUTPUT_FAILURE_RATIO_KEY = ".noise.output.failure.p";
	private static final String NOISE_OUTPUT_DROPOUT_USE_KEY = ".noise.output.dropout.use";
	private static final String NOISE_OUTPUT_DROPOUT_RATIO_KEY = ".noise.output.dropout.p";

    private static final String TIMESTEP_KEY = "kernel.timesteps";
	private static final int DEFAULT_TIMESTEP = 300;
	////////////////////////////////////////////////////////////////////////////////////////


	private final Random random;

	private final AdvancedWorldModel world;

	/**
	 * 使用可能なチャンネル数．
	 */
	protected int channelCount;
	/**
	 * 送信するチャネル．
	 */
	protected int sendChannel;
	/**
	 * 受信するチャネル．
	 */
	protected int[] subscribeChannels;
	/**
	 * Platoonがsubscribe（受信）可能なチャンネル数．
	 */
	protected int subscribePlatoonSize;
	/**
	 * Centerがsubscribe（受信）可能なチャンネル数．
	 */
	protected int subscribeCenterSize;
	/**
	 * 使用可能な無線チャネル情報のリスト．
	 */
	protected List<AbstractChannel> radioChannels;
	/**
	 * 使用可能な声チャネル情報のリスト．
	 */
	protected List<AbstractChannel> voiceChannels;
	/**
	 * 全radio channelの帯域幅合計．
	 */
	protected int bandwidthAll;
	/**
	 * タイム毎の乱数値．
	 */
	private int[] randList;

	public ChannelManager(CommunicationAgent agent, AdvancedWorldModel world, Config config) {
		this.world = world;
		random = new Random(agent.getID().getValue());
		channelCount = config.getIntValue(CHANNELS_COUNT_KEY);
		bandwidthAll = 0;
		//-----------------
		//チャネル情報の設定
		//-----------------
		radioChannels = new ArrayList<AbstractChannel>();
		voiceChannels = new ArrayList<AbstractChannel>();
		for (int i=0; i<channelCount; i++) {
			AbstractChannel channel = null;
			String type = config.getValue(CHANNELS_KEY_PREFIX + i + CHANNELS_TYPE_KEY);
			if (type.startsWith("voice")) {
				int range = config.getIntValue(CHANNELS_KEY_PREFIX + i + RANGE_SUFFIX);
				int size = config.getIntValue(CHANNELS_KEY_PREFIX + i + MESSAGE_SIZE_SUFFIX);
				int number = config.getIntValue(CHANNELS_KEY_PREFIX + i + MESSAGE_MAX_SUFFIX);
				channel = new VoiceChannel(i, range, size, number);
				voiceChannels.add(channel);
			} else if (type.startsWith("radio")) {
				int bandwidth = config.getIntValue(CHANNELS_KEY_PREFIX + i + RADIO_BAND_WIDTH_KEY);
				RadioChannel radio = new RadioChannel(i, bandwidth);
				bandwidthAll += radio.getBandwidth();
				channel = radio;
				radioChannels.add(channel);
			}
//			//noise関係の設定
//			if (config.getBooleanValue(CHANNELS_KEY_PREFIX + i + NOISE_INPUT_FAILURE_USE_KEY, false)) {
//				FailureNoise noise = new FailureNoise();
//				noise.useNoise();
//				noise.setRatio(config.getFloatValue(CHANNELS_KEY_PREFIX + i + NOISE_INPUT_FAILURE_RATIO_KEY, 0.0));
//				channel.addInputNoise(noise);
//			}
//			if (config.getBooleanValue(CHANNELS_KEY_PREFIX + i + NOISE_INPUT_DROPOUT_USE_KEY, false)) {
//				DropoutNoise noise = new DropoutNoise();
//				noise.useNoise();
//				noise.setRatio(config.getFloatValue(CHANNELS_KEY_PREFIX + i + NOISE_INPUT_DROPOUT_RATIO_KEY, 0.0));
//				channel.addInputNoise(noise);
//			}
//			if (config.getBooleanValue(CHANNELS_KEY_PREFIX + i + NOISE_OUTPUT_FAILURE_USE_KEY, false)) {
//				FailureNoise noise = new FailureNoise();
//				noise.useNoise();
//				noise.setRatio(config.getFloatValue(CHANNELS_KEY_PREFIX + i + NOISE_OUTPUT_FAILURE_RATIO_KEY, 0.0));
//				channel.addOutputNoise(noise);
//			}
//			if (config.getBooleanValue(CHANNELS_KEY_PREFIX + i + NOISE_OUTPUT_DROPOUT_USE_KEY, false)) {
//				DropoutNoise noise = new DropoutNoise();
//				noise.useNoise();
//				noise.setRatio(config.getFloatValue(CHANNELS_KEY_PREFIX + i + NOISE_OUTPUT_DROPOUT_RATIO_KEY, 0.0));
//				channel.addOutputNoise(noise);
//			}
		}
		//------------------------
		//受信可能なチャネル数の設定
		//------------------------
		subscribePlatoonSize = config.getIntValue(SUBSCRIBE_PLATOON_KEY, 0);
		//無いと思うが受信可能回数が全チャネル数を上回る場合，全チャネル数を上限とする．
		subscribePlatoonSize = (subscribePlatoonSize > channelCount) ? channelCount : subscribePlatoonSize;
		//0のときの処理を作る
		subscribeChannels = new int[subscribePlatoonSize];
		subscribeCenterSize = config.getIntValue(SUBSCRIBE_CENTER_KEY);
		//----------------------
		//チャネル選択用乱数の生成
		//----------------------
		int timestep = agent.getTimestep();
		randList = new int[timestep];
		//java.util.Random.nextInt()の引数に0以下が入る時の処理．
		if (0 < bandwidthAll) {
			for (int i = 0; i < timestep; i++) {
				randList[i] = random.nextInt(bandwidthAll);
			}
		} else {
			for (int i = 0; i < timestep; i++) {
				randList[i] = random.nextInt();
			}
		}
	}


	/**
	 * チャネル番号からチャネル情報を取得
	 * @param c
	 * @return
	 */
	public AbstractChannel getChannel(int c) {
		for (AbstractChannel ch : voiceChannels) {
			if (ch.getChannel() == c) {
				return ch;
			}
		}
		for (AbstractChannel ch : radioChannels) {
			if (ch.getChannel() == c) {
				return ch;
			}
		}
		return null;
	}

	/**
	 * 送信するチャネルを選択する．
	 * @return 送信チャネル番号．
	 */
	public int getSendChannel() {
		sendChannel = getRandomChannel(world.time-1, 0);
		return sendChannel;
	}

	/**
	 * 受信するチャネルを選択する．
	 * @return 受信チャネル番号．
	 */
	public int[] getSubscribeChannels() {
		for (int i = 0; i < subscribeChannels.length; i++) {
			subscribeChannels[i] = getUnsubscribeChannel(i);
		}
		return subscribeChannels;
	}

	/**
	 * 受信用でまだ使っていないチャネルを選択する．
	 * @param useNum 選択済み受信チャネルの数．
	 * @return 未選択受信チャネル．
	 */
	private int getUnsubscribeChannel(int useNum) {
		int result = radioChannels.get(0).getChannel();
		boolean determined = false;
		if (useNum < 0) {
			useNum *= -1;
		}
		useNum %= subscribeChannels.length;

		for (int limit = 1000; !determined && 0 < limit; limit--) {
			determined = true;
			result = getRandomChannel(random.nextInt(bandwidthAll), 0);
			for (int i = 0; i < useNum; i++) {
				if (result == subscribeChannels[i]) {
					determined = false;
				}
			}
		}
		return result;
	}

	/**
	 * 使用するチャネルをランダムに選択する
	 * @param randIndex 使用する乱数．
	 * @param offset 乱数の調整用．
	 * @return 使用するチャネル．
	 */
	private int getRandomChannel(int randIndex, int offset) {
		int result = ((RadioChannel)radioChannels.get(0)).getChannel();
		if (randIndex < 0) {
			randIndex *= -1;
		}
		randIndex %= randList.length;

		for (int rand = (randList[randIndex] + offset) % bandwidthAll, i = 0; 0 <= rand; i++) {
			RadioChannel rc = (RadioChannel)radioChannels.get(i);
			result = rc.getChannel();
			rand -= rc.getBandwidth();
		}
		return result;
	}

	/**
	 * radio channelの帯域幅の合計を返す．
	 * @return radio channelの帯域幅の合計．
	 */
	public int getBandwidthAll() {
		return bandwidthAll;
	}

	/**
	 * radio channelの数を返す．
	 * @return radio channelの数．
	 */
	public int getRadioChannelsSize() {
		return radioChannels.size();
	}

}
