package rione.object;

import java.util.*;

import rescuecore2.config.Config;
import rescuecore2.misc.Pair;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.registry.Registry;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.worldmodel.AbstractEntity;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.EntityListener;
import rescuecore2.worldmodel.Property;
import rione.agent.AgentConstants;
import rione.util.UpdateHashSet;

/**
 * 災害空間の情報を保持するクラス
 */
public class AdvancedWorldModel extends StandardWorldModel {

	/** Streetのリスト */
	private ArrayList<Street> streets;

	/** あるEntityの属するStreet */
	private final Map<EntityID, Street> ownStreet = new HashMap<EntityID, Street>();

	/** 孤立しており、行くことができないEntityのリスト */
	private Set<StandardEntity> isolatedEntities;

	/** 自分を表す */
	private Human me;

	/** EntityIDから背番号を得るMap */
	private Map<EntityID, Integer> idToUniform;

	/** URNと背番号からEntityIDを得るMap */
	private Map<String, Map<Integer, EntityID>> uniformToID;

	private Map<Building, Set<Road>> entrance = new HashMap<Building, Set<Road>>();

	/** Worldに保存されているFireArea */
	private ArrayList<FireArea> fireAreas = new ArrayList<FireArea>();

	/**
	 * 最後に情報を更新した時間
	 */
	private Map<EntityID, Map<StandardPropertyURN, Integer>> updatedPropertyTimeStamp = new LazyMap<EntityID, Map<StandardPropertyURN, Integer>>() {
		@Override
		public Map<StandardPropertyURN, Integer> createValue() {
			return new HashMap<StandardPropertyURN, Integer>();
		}
	};

	/** 最後に状態が変化した時間 */
	private Map<EntityID, Integer> lastChangedTime = new HashMap<EntityID, Integer>();

	/** 最後に見た時間 */
	private Map<EntityID, Integer> lastSeenTime = new HashMap<EntityID, Integer>();

	/** 時間を保持する */
	public int time;

	/** 終了時間 */
	public int timestep = 300;

	/**
	 * 燃えている建物の集合
	 */
	private Set<Building> burningBuildings = new HashSet<Building>();

	private ReachableArea reachable;

	/** 閉塞の中にいるエージェントの集合 */
	private Set<Human> locateInBlockadeAgents = new HashSet<Human>();

	protected double woodenIgnition;
	protected double steelIgnition;
	protected double concreteIgnition;

	/**
	 * コンストラクタ
	 */
	public AdvancedWorldModel() {
		super();
		// Streetを作成する.
		// streets = Street.createStreet(this);
	}

	/**
	 * WorldModelを初期化する
	 */
	public void initialize(int timestep, Config config) {
		for (StandardEntity entity : this) {
			entity.addEntityListener(new StateChangeListener());
			if (entity instanceof Road) {
				entity.addEntityListener(new AreaBlockadeListener());
			}
		}
		setEntrance();
		
		final double DEFAULT_WOODEN_IGNITION = 47.0f;
		final double DEFAULT_STEEL_IGNITION = 47.0f;
		final double DEFAULT_CONCRETE_IGNITION = 47.0f;

		final String WOODEN_IGNITION_KEY = "resq-fire.wooden_ignition";
		final String STEEL_IGNITION_KEY = "resq-fire.steel_ignition";
		String CONCRETE_IGNITION_KEY = "resq-fire.concrete_ignition";
		woodenIgnition = config.getFloatValue(WOODEN_IGNITION_KEY, DEFAULT_WOODEN_IGNITION);
		steelIgnition = config.getFloatValue(STEEL_IGNITION_KEY, DEFAULT_STEEL_IGNITION);
		concreteIgnition = config.getFloatValue(CONCRETE_IGNITION_KEY, DEFAULT_CONCRETE_IGNITION);
	}

	/**
	 * worldmodelを更新
	 * 
	 * @param changed
	 */
	public void update(ChangeSet changed) {
		// 見た場所にCivilianが居ても削除されないので手動で消す
		for (Iterator<StandardEntity> it = getEntitiesOfType(StandardEntityURN.CIVILIAN).iterator(); it.hasNext();) {
			Civilian cv = (Civilian) it.next();
			if (me().getPosition(this).equals(cv.getPosition(this)) && !changed.getChangedEntities().contains(cv.getID())) {
				it.remove();
			}
		}
		
		predictBurnedOut();
		
		updateFireAreas();
		updateLocateInBlockade(changed);
		needRescueHuman.update();
	}

	public List<Building> determinatedBurnedOut = new ArrayList<Building>();
	public HashSet<Building> maybeBurnedOut = new HashSet<Building>(3000);
	private void predictBurnedOut() {
		if (time % 3 != me().getID().getValue() % 3) return; 
		maybeBurnedOut.clear();
		determinatedBurnedOut.clear();
		
		for (StandardEntity se : getEntitiesOfType(StandardEntityURN.BUILDING)) {
			Building bld = (Building) se;
			if (bld.isFierynessDefined() && bld.getFierynessEnum().equals(Fieryness.BURNT_OUT)) {
				determinatedBurnedOut.add(bld);
				maybeBurnedOut.add(bld);
			}
		}
		
		for (Building bld : determinatedBurnedOut) {
			ArrayDeque<Pair<Building, Integer>> q = new ArrayDeque<Pair<Building, Integer>>();
			q.offerFirst(new Pair<Building, Integer>(bld, 0));
			while (!q.isEmpty()) {
				Pair<Building, Integer> current = q.pollLast();
				if (current.second() >= 1 + Math.min(time, 300) / 50) continue;
				int addCount = 0;
				NEXT_NODE_LOOP : for (StandardEntity next :
						getObjectsInRange(
								current.first(),
								(int) (bld.isGroundAreaDefined() ? Math.sqrt(bld.getGroundArea()) : 24000) + 20000)) {
					if (next instanceof Building) {
						Building nextBld = (Building) next;
						if (nextBld instanceof AmbulanceCentre
								|| nextBld instanceof FireStation
								|| nextBld instanceof PoliceOffice
								|| nextBld instanceof Refuge) 
							 continue NEXT_NODE_LOOP;
						if (maybeBurnedOut.contains(nextBld)) continue NEXT_NODE_LOOP;
						int lastChanged = getLastChangedTime(nextBld.getID());
						if (lastChanged != -1) {
							int notChangedTime = time - lastChanged;
							if (nextBld.isFierynessDefined()) {
								switch (nextBld.getFierynessEnum()) {
								case HEATING:
									if (notChangedTime < 14) {
										continue NEXT_NODE_LOOP;
									}
									break;
								case BURNING:
									if (notChangedTime < 5) {
										continue NEXT_NODE_LOOP;
									}
									break;
								case INFERNO:
									if (notChangedTime < 3) {
										continue NEXT_NODE_LOOP;
									}
									break;
								default:
									continue NEXT_NODE_LOOP;
								}
							}
							else if (notChangedTime < 3) {
								continue NEXT_NODE_LOOP;
							}
						}
						
						
						maybeBurnedOut.add(nextBld);
						q.offerFirst(new Pair<Building, Integer>(nextBld, current.second() + 1));
						addCount++;
						if (addCount == 4) break;
					}
				}
			}
		}
	}

	/**
	 * Streetのリストを返す.
	 * 
	 * @return
	 */
	public ArrayList<Street> streets() {
		return streets;
	}

	/**
	 * あるEntityの所属するStreetを設定する. Streetクラスから呼び出す.
	 * 
	 * @param id
	 * @param street
	 */
	protected void setOwnStreet(EntityID id, Street street) {
		ownStreet.put(id, street);
	}

	/**
	 * あるEntityの所属するStreetを返す.
	 * 
	 * @param id
	 * @return
	 */
	public Street getOwnStreet(EntityID id) {
		return ownStreet.get(id);
	}

	/**
	 * あるRoadが、Streetの始点,終点になるかどうかを返す.
	 * 
	 * @param road
	 *                注目するRoad
	 * @return 始点,終点になるならtrue, ならないならfalse
	 */
	protected boolean isStreetRoad(Road road) {
		// roadに隣接しているRoadの数を数える.
		int counter = 0;
		for (EntityID id : road.getNeighbours()) {
			StandardEntity entity = getEntity(id);
			if (entity instanceof Road) {
				counter++;
			}
		}
		// 隣接しているRoadが2つでないときtrue
		return (counter != 2) ? true : false;
	}

	/**
	 * 孤立しており、行くことができないEntityを設定する. Street作成時に呼び出す.
	 * 
	 * @param col
	 *                孤立していないEntityのコレクション
	 */
	protected void setIsolationEntities(Collection<StandardEntity> col) {
		// Set<StandardEntity> allEntities = new
		// HashSet<StandardEntity>(getAllEntities());
		// allEntities.removeAll(getEntitiesOfType(AgentConstants.HUMANOIDS));
		// allEntities.removeAll(col);
		// isolationEntities = allEntities;
	}

	/**
	 * 孤立したAreaのセットを設定する Streetを作成していない現状の一時的な処理
	 */
	public void setIsolationEntities() {
		isolatedEntities = new HashSet<StandardEntity>(getEntitiesOfType(AgentConstants.AREAS));
		Area area = (Area) getEntity(me().getPosition());
		setIsolationEntities(area);
	}

	/**
	 * 再帰的につながっているAreaを孤立しているAreaのSetから除外する
	 */
	private void setIsolationEntities(Area area) {
		if (!isolatedEntities.contains(area)) {
			return;
		}
		isolatedEntities.remove(area);
		List<EntityID> neighbors = area.getNeighbours();
		for (EntityID id : neighbors) {
			Area a = (Area) getEntity(id);
			setIsolationEntities(a);
		}
	}

	/**
	 * 再帰的につながっているAreaを孤立しているAreaのSetから除外する
	 */
	public boolean isIsolation(EntityID position) {
		try {
			return isIsolation((Area) getEntity(position));
		}
		catch (ClassCastException e) {
			return false;
		}
	}

	/**
	 * 孤立しており、行くことができない場所のセットを返す。
	 * 
	 * @return 孤立しているAreaのセット
	 */
	public Set<StandardEntity> getIsolationEntities() {
		return isolatedEntities;
	}

	/**
	 * 孤立しており、行くことができない場所であるかどうか。
	 * 
	 * @param area
	 *                判定するArea
	 * @return 行くことができないならばtrue
	 */
	public boolean isIsolation(Area area) {
		return isolatedEntities.contains(area);
	}

	/**
	 * WorldModelに自分の情報を登録する
	 * 
	 * @param me
	 *                自分
	 */
	public void setMe(Human me) {
		this.me = me;
	}

	/**
	 * 自分の情報を返す
	 * 
	 * @return 自分
	 */
	public Human me() {
		return me;
	}

	/**
	 * IDと背番号の相互変換のMapを登録する
	 * 
	 * @param idToUniform
	 * @param uniformToID
	 */
	public void setUniform(Map<EntityID, Integer> idToUniform, Map<String, Map<Integer, EntityID>> uniformToID) {
		this.idToUniform = idToUniform;
		this.uniformToID = uniformToID;
	}

	/**
	 * EntityIDから背番号を得る。 背番号の存在しないEntityの場合は-1を返す。
	 * 
	 * @param id
	 *                背番号を得たいEntityのID
	 * @return 背番号
	 */
	public int getUniform(EntityID id) {
		Integer uniform = idToUniform.get(id);
		if (uniform == null) {
			return -1;
		}
		return uniform.intValue();
	}

	/**
	 * StandatdEntityから背番号を得る。 背番号の存在しないEntityの場合は-1を返す。
	 * 
	 * @param entity
	 *                背番号を得たいEntity
	 * @return 背番号
	 */
	public int getUniform(StandardEntity entity) {
		return getUniform(entity.getID());
	}

	/**
	 * URNと背番号からEntityIDを得る
	 * 
	 * @param urn
	 *                Entityを表すURN
	 * @param uniform
	 *                背番号
	 * @return
	 */
	public EntityID getEntityIDfromUniform(StandardEntityURN urn, int uniform) {
		if (!AgentConstants.UNIFORMABLE_ENTITIES.contains(urn)) {
			return null;
		}
		return uniformToID.get(urn.toString()).get(uniform);
	}

	/**
	 * URNからStreetを含むEntityの集合を得る
	 * 
	 * @param urn
	 *                URN
	 * @return Entityの集合
	 */
	public Collection<? extends AbstractEntity> getEntitiesOfType(String urn) {
		if (urn.equals(RioneEntityURN.STREET.toString())) {
			return streets;
		}
		return getEntitiesOfType(StandardEntityURN.fromString(urn));
	}

	/**
	 * EnumSetからEntityの集合を得る
	 * 
	 * @param urns
	 *                URNのセット
	 * @return Entityの集合
	 */
	public Collection<StandardEntity> getEntitiesOfType(EnumSet<StandardEntityURN> urns) {
		Collection<StandardEntity> res = new HashSet<StandardEntity>();
		for (StandardEntityURN urn : urns) {
			res.addAll(getEntitiesOfType(urn));
		}
		return res;
	}

	/**
	 * X, Y座標から場所を得る
	 * 
	 * @param x
	 *                X座標
	 * @param y
	 *                Y座標
	 * @return 場所
	 */
	public Area getPositionFromCoordinates(int x, int y) {
		for (StandardEntity entity : getObjectsInRange(x, y, 1)) {
			if (entity instanceof Area) {
				Area area = (Area) entity;
				if (area.getShape().contains(x, y)) {
					return area;
				}
			}
		}
		return null;
	}

	/**
	 * 座標から場所を得る
	 * 
	 * @param coordinate
	 *                座標
	 * @return 場所
	 */
	public StandardEntity getPositionFromCoordinates(Pair<Integer, Integer> coordinate) {
		return getPositionFromCoordinates(coordinate.first(), coordinate.second());
	}

	/**
	 * EntityのPropertyごとの最終更新情報を得る
	 * 
	 * @param id
	 * @return
	 */
	public Map<StandardPropertyURN, Integer> getPropertyTimeStampMap(EntityID id) {
		return updatedPropertyTimeStamp.get(id);
	}

	/**
	 * Propertyの最終更新時間を得る
	 * 
	 * @param id
	 * @param propertyUrn
	 * @return
	 */
	public int getPropertyTimeStamp(EntityID id, StandardPropertyURN propertyUrn) {
		Integer time = updatedPropertyTimeStamp.get(id).get(propertyUrn);
		return time != null ? time : -1;
	}

	/**
	 * Propertyの更新時間をセットする
	 * 
	 * @param id
	 * @param propertyUrn
	 * @param timeStamp
	 */
	public void setProtertyTimestamp(EntityID id, StandardPropertyURN propertyUrn, int timeStamp) {
		Map<StandardPropertyURN, Integer> map = updatedPropertyTimeStamp.get(id);
		map.put(propertyUrn, timeStamp);
	}

	/**
	 * Propertyの更新時間を現在時刻にセットする
	 * 
	 * @param id
	 * @param propertyUrn
	 */
	public void setProtertyTimestamp(EntityID id, StandardPropertyURN propertyUrn) {
		setProtertyTimestamp(id, propertyUrn, time);
	}

	/**
	 * そのEntityを最後に見た時間<br>
	 * 見ていない場合は-1
	 * 
	 * @param id
	 * @return
	 */
	public int getLastSeenTime(EntityID id) {
		Integer time = lastSeenTime.get(id);
		return (time != null) ? time : -1;
	}

	/**
	 * そのEntityを最後に見た時間を設定する
	 * 
	 * @param id
	 * @param timeStamp
	 */
	private void setLastSeenTime(EntityID id, int timeStamp) {
		lastSeenTime.put(id, timeStamp);
	}

	/**
	 * Entityごとの更新時間を現在時刻にセットする
	 * 
	 * @param id
	 */
	public void setEntityTimestamp(EntityID id) {
		setLastSeenTime(id, time);
	}

	/**
	 * 燃えている建物をセットする
	 * 
	 * @param building
	 */
	public void setBurningBuildings(Building building) {
		if (building.isOnFire()) {
			burningBuildings.add(building);
		}
		else {
			burningBuildings.remove(building);
		}
	}

	/**
	 * 燃えている建物の集合
	 * 
	 * @return
	 */
	public Set<Building> getBurningBuildings() {
		return burningBuildings;
	}

	/**
	 * AbstractWorldModeのメソッドをオーバーライドし 情報の更新時にタイムスタンプを更新するようにしている
	 */
	@Override
	public void merge(ChangeSet changeSet) {
		for (EntityID e : changeSet.getChangedEntities()) {
			StandardEntity existingEntity = getEntity(e);
			if (existingEntity == null) {
				// 新しくEntityを作成してWorldModelに登録する
				existingEntity = (StandardEntity) Registry.getCurrentRegistry().createEntity(changeSet.getEntityURN(e), e);
				existingEntity.addEntityListener(new StateChangeListener());
				addEntity(existingEntity);
			}
			Map<StandardPropertyURN, Integer> propertyTimestampMap = getPropertyTimeStampMap(existingEntity.getID());
			if (propertyTimestampMap == null) {
				propertyTimestampMap = new HashMap<StandardPropertyURN, Integer>();
			}
			for (Property p : changeSet.getChangedProperties(e)) {
				Property existingProperty = existingEntity.getProperty(p.getURN());
				existingProperty.takeValue(p);
				// Propertyのタイムスタンプを更新
				propertyTimestampMap.put(StandardPropertyURN.fromString(p.getURN()), time);
			}
			// Entityのタイムスタンプを更新
			setLastSeenTime(existingEntity.getID(), time);

			// burningBuidnigの設定
			if (existingEntity instanceof Building) {
				setBurningBuildings((Building) existingEntity);
			}
		}
		for (EntityID next : changeSet.getDeletedEntities()) {
			removeEntity(next);
		}
	}

	/**
	 * Entityの状態が変化したことを監視するListener
	 * 
	 * @author com
	 * 
	 */
	private class StateChangeListener implements EntityListener {
		@Override
		public void propertyChanged(Entity entity, Property property, Object oldValue, Object newValue) {
			lastChangedTime.put(entity.getID(), time);
		}

	}

	/**
	 * EntityにStateChangeListenerを付加する 通信情報で初めて認識したEntityに使用する。
	 * 
	 * @param entity
	 *                StateChangeListenerを付加するEntity
	 */
	public void addStateChangeListener(Entity entity) {
		entity.addEntityListener(new StateChangeListener());
	}

	/**
	 * Area情報が更新された際、啓開済みの閉塞をWorldModelから削除するリスナー
	 * 
	 * @author com
	 * 
	 */
	private class AreaBlockadeListener implements EntityListener {

		@SuppressWarnings("unchecked")
		@Override
		public void propertyChanged(Entity entity, Property property, Object oldValue, Object newValue) {
			if (property.getURN().equals(StandardPropertyURN.BLOCKADES.toString())) {
				if (oldValue == null || newValue == null) {
					return;
				}
				List<EntityID> before = (List<EntityID>) oldValue;
				List<EntityID> after = (List<EntityID>) newValue;
				before.removeAll(after);
				for (EntityID id : before) {
					removeEntity(id);
				}
			}
		}
	}

	/**
	 * Entityの状態が最後に変化した時間を得る
	 * 
	 * @param id
	 *                変化した時間を得たいEntity
	 * @return Entityの状態が最後に変化した時間
	 */
	public int getLastChangedTime(EntityID id) {
		Integer t = lastChangedTime.get(id);
		return t != null ? t.intValue() : -1;
	}

	/**
	 * 座標が閉塞の内側かどうかを返す。
	 * 
	 * @param x
	 *                判定するX座標
	 * @param y
	 *                判定するY座標
	 * @return 閉塞の内側ならtrue
	 */
	public boolean isLocateInBlockade(int x, int y) {
		// TODO WorldModelから削除されていない実態のない閉塞で判定される可能性がある。
		for (StandardEntity entity : getObjectsInRange(x, y, 1)) {
			if (!(entity instanceof Blockade)) {
				continue;
			}
			Blockade blockade = (Blockade) entity;
			if (blockade.getShape().contains(x, y)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 座標が閉塞の内側かどうかを返す。
	 * 
	 * @param location
	 *                判定する座標
	 * @return 閉塞の内側ならtrue
	 */
	public boolean isLocateInBlockade(Pair<Integer, Integer> location) {
		return isLocateInBlockade(location.first(), location.second());
	}

	/**
	 * Humanが閉塞の内側にいるかどうかを返す。
	 * 
	 * @param human
	 *                判定するHuman
	 * @return 閉塞の内側にいるならtrue
	 */
	public boolean isLocateInBlockade(Human human) {
		return isLocateInBlockade(human.getLocation(this));
	}

	/**
	 * 閉塞内にHumanがいるかどうかを返す。
	 * 
	 * @param blockade
	 *                判定する閉塞
	 * @return 内部にHumanがいるときtrue、いないときfalse
	 */
	public boolean isContainsHuman(Blockade blockade) {
		if (!blockade.isApexesDefined()) {
			return false;
		}
		for (StandardEntity entity : getEntitiesOfType(AgentConstants.HUMANOIDS)) {
			Pair<Integer, Integer> location = entity.getLocation(this);
			if (blockade.getShape().contains(location.first(), location.second())) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 建物の入り口を設定する。
	 */
	private void setEntrance() {
		for (StandardEntity entity : getEntitiesOfType(AgentConstants.BUILDINGS)) {
			Building building = (Building) entity;
			entrance.put((Building) entity, setEntrance(building));
		}
	}

	/**
	 * 建物の入り口を設定する。<br>
	 * 建物と直接接続している場合を考慮し、 深さ優先で探索する。
	 * 
	 * @param building
	 *                入り口を設定する建物
	 * @return 建物の入り口の集合
	 */
	private Set<Road> setEntrance(Building building) {
		Set<Road> roads = new HashSet<Road>();
		Stack<StandardEntity> stack = new Stack<StandardEntity>();
		Set<StandardEntity> visited = new HashSet<StandardEntity>();
		stack.add(building);
		do {
			StandardEntity entity = stack.pop();
			visited.add(entity);
			if (entity instanceof Road) {
				roads.add((Road) entity);
			}
			else if (entity instanceof Building) {
				Building bld = (Building) entity;
				for (EntityID id : bld.getNeighbours()) {
					StandardEntity e = getEntity(id);
					if (!visited.contains(e)) {
						stack.push(e);
					}
				}
			}
		} while (!stack.isEmpty());
		return roads;
	}

	/**
	 * 建物の入り口のRoadを得る。<br>
	 * 複数の入り口がある場合を考慮し、Setで返す。
	 * 
	 * @param building
	 *                入り口を得る建物
	 * @return 入り口の集合
	 */
	public Set<Road> getEntrance(Building building) {
		return entrance.get(building);
	}

	public ReachableArea getReachable() {
		return reachable;
	}

	public void setReachable(ReachableArea reachable) {
		this.reachable = reachable;
	}

	public Set<Human> getLocateInBlockadeAgents() {
		return locateInBlockadeAgents;
	}

	/**
	 * 閉塞の中にいるエージェントを集合に追加する。
	 * 
	 * @param hu
	 *                閉塞の中にいるエージェント
	 */
	public void addLocateInBlockadeAgents(Human hu) {
		locateInBlockadeAgents.add(hu);
	}

	/**
	 * 閉塞の中にいるエージェントを集合から削除する。
	 * 
	 * @param hu
	 *                削除するエージェント
	 */
	public void removeLocateInBlockadeAgents(Human hu) {
		locateInBlockadeAgents.remove(hu);
	}

	/**
	 * Worldの中にあるFireAreaを更新する
	 */
	private void updateFireAreas() {
		// update (expand)
		for (Iterator<FireArea> it = fireAreas.iterator(); it.hasNext();) {
			FireArea fa = it.next();
			fa.update();
			// FireArea拡大
			HashSet<Building> nears = new HashSet<Building>(fa.getNears());
			HashSet<Building> newNears = new HashSet<Building>(nears);
			for (Building bld : nears) {
				// 燃えている場合はさらに周囲の建物を追加する
				if (bld.isOnFire()) {
					// 燃えている建物として新に追加
					fa.addBurningBuilding(bld);
					int length = bld.isGroundAreaDefined() ? (int) Math.sqrt(bld.getGroundArea()) * 1000 * 2 : 20 * 1000;
					newNears.remove(bld);
					// 周囲の建物を新にnearに追加
					for (Entity e : this.getObjectsInRange(bld, length)) {
						if (e instanceof Building) {
							Building b = (Building) e;
							newNears.add(b);
						}
					}
				}
			}
			fa.setNears(newNears);
			// 消火しきった（燃え尽きた）ときのremove処理
			if (!fa.isOnFire()) {
				it.remove();
			}
		}
		// unionFireAreas
		// FireGenealogy が重なったら1つに統合する
		for (int i = 0; i < this.fireAreas.size() - 1; i++) {
			FireArea fa1 = this.fireAreas.get(i);
			for (int j = i + 1; j < this.fireAreas.size(); j++) {
				FireArea fa2 = this.fireAreas.get(j);
				if (fa2.unionFireArea(fa1)) {
					this.fireAreas.remove(i--);
					break;
				}
			}
		}
		// どの火災エリアにも属しない火災から新に火災エリアを作る
		// 現在の火災
		HashSet<Building> burnings = new HashSet<Building>(this.getBurningBuildings());
		// すでにどこかのFireAreaに追加した火災
		HashSet<Building> finished = new HashSet<Building>();

		// すでにFireAreaに含まれているもの
		for (FireArea fa : this.fireAreas) {
			finished.addAll(fa.getBuildings());
		}

		for (Building bld : burnings) {
			// すでに見たものは無視
			if (finished.contains(bld)) {
				continue;
			}
			// 1つの建物に隣接しているとみなす距離
			int length = bld.isGroundAreaDefined() ? (int) Math.sqrt(bld.getGroundArea()) * 1000 * 2 : 20 * 1000;
			// 1つの建物に対し、一定範囲内の燃えている建物は1つのグループとみなす
			HashSet<Building> aroundBurnings = new HashSet<Building>();
			// 一定距離内の建物をすべて確認
			for (Entity e : this.getObjectsInRange(bld, length)) {
				if (!(e instanceof Building)) {
					continue;
				}
				Building b = (Building) e;
				if (!finished.contains(b)) {
					aroundBurnings.add(b);
				}
			}
			aroundBurnings.add(bld);

			// 1つのグループとみなす建物群を中心として新しいFireAreaを生成する
			fireAreas.add(new FireArea(aroundBurnings));
			finished.addAll(aroundBurnings);
		}
	}

	/**
	 * Worldに現存するFireAreaを返す
	 * 
	 * @return
	 */
	public List<FireArea> getFireAreas() {
		return this.fireAreas;
	}

	/**
	 * 閉塞の中にいるエージェント情報を更新する。
	 */
	private void updateLocateInBlockade(ChangeSet changed) {
		for (EntityID id : changed.getChangedEntities()) {
			StandardEntity entity = getEntity(id);
			switch (entity.getStandardURN()) {
			case FIRE_BRIGADE:
			case AMBULANCE_TEAM:
			case POLICE_FORCE:
				if (isLocateInBlockade((Human) entity)) {
					addLocateInBlockadeAgents((Human) entity);
				}
				break;
			default:
				break;
			}
		}
	}

	/**
	 * Buildingの中に居るCivilianもしくはBuildingの中に居る埋まっているエージェントの集合
	 */
	public UpdateHashSet<EntityID> needRescueHuman = new UpdateHashSet<EntityID>() {
		private static final long serialVersionUID = 1L;

		@Override
		public void update() {
			clear();
			for (StandardEntity se : getEntitiesOfType(AgentConstants.HUMANOIDS)) {
				add((Human) se);
			}
		}

		public boolean add(Human human) {
			if (human.equals(me()))
				return false;
			if (human.isHPDefined()) {
				if (human.getHP() <= 0) {
					return false;
				}
				else if (human.isDamageDefined()) {
					int deathTime = getDeathTime(human);
					if (0 <= deathTime && deathTime <= time + 3) {
						// 3は「積む・運ぶ・降ろす」のに最低限必要な時間
						return false;
					}
				}
			}
			if (human.isBuriednessDefined()) {
				if (!(human instanceof Civilian) && human.getBuriedness() == 0) {
					return false;
				}
			}
			if (human.isPositionDefined()) {
				StandardEntity area = getEntity(human.getPosition());
				if (area instanceof Building 
						&& !(area instanceof Refuge)
						&& !maybeBurnedOut.contains(area)) {
					if (human instanceof Civilian) {
						return super.add(human.getID());
					}
					else if (human.isBuriednessDefined() && human.getBuriedness() > 0) {
						return super.add(human.getID());
					}
				}
			}
			return false;
		}
	};

	public void addHeardCivilian(Civilian civilian) {
		if (needRescueHuman.add(civilian.getID()) && getEntity(civilian.getID()) == null) {
			addStateChangeListener(civilian);
			addEntity(civilian);
		}
	}

	/**
	 * 死亡予測時刻<br>
	 * ダメージがない場合はtimestep + 1<br>
	 * 判定できない場合は負<br>
	 * timestep以上に生きる場合もある<br>
	 * 使う場合は要注意<br>
	 * Damageの変化を考慮していないので，さらに早く死ぬ可能性がある
	 * 
	 * @param human
	 * @return
	 */
	public int getDeathTime(Human human) {
		if (human.isDamageDefined()) {
			if (human.getDamage() == 0) {
				return Integer.MAX_VALUE / 2;
			}
			else {
				int lastSeenTime = getLastSeenTime(human.getID());
				if (lastSeenTime >= 0 && human.isHPDefined()) {
					return lastSeenTime + human.getHP() / human.getDamage();
				}
				else {
					return -1;
				}
			}
		}
		return -1;
	}
}
