package rione.object;

import java.util.HashSet;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;

public class FireArea implements Comparable<FireArea>{
	/** このグループに含まれる建物 */
	private HashSet<Building> buildings = new HashSet<Building>();
	/** このグループの中で燃えている建物 */
	private HashSet<Building> burningBuildings = new HashSet<Building>();
	/** このグループの中で燃え尽きた建物 */
	private HashSet<Building> burntOutBuildings = new HashSet<Building>();
	/** このグループの近くで燃えそうな建物 */
	private HashSet<Building> nears = new HashSet<Building>();
	/** このグループの発火点となった建物 */
	private HashSet<Building> ignitions = new HashSet<Building>();
	
	/**
	 * 燃えている建物の集合からグループを作るコンストラクタ
	 * @param burnings
	 */
	public FireArea(HashSet<Building> buildings){
		// すべて所属建物群に追加
		this.buildings.addAll(buildings);
		for(Building bld : buildings){
			// 各建物について燃えていればburningに追加
			// そうでなければnearsに追加
			if(bld.isOnFire()){
				this.burningBuildings.add(bld);
				this.ignitions.add(bld);
			}
			else{
				this.nears.add(bld);
			}
		}
		this.update();
	}
	
	/**
	 * 燃えている建物からグループを作るコンストラクタ
	 * @param burningBuilding
	 */
	public FireArea(Building burningBuilding){
		this.buildings.add(burningBuilding);
		this.burningBuildings.add(burningBuilding);
		this.ignitions.add(burningBuilding);
		this.update();
	}
	
	/**
	 * このグループ内に実際のWorldの建物の情報を反映させる
	 */
	public void update(){
		// 燃えている建物が燃え尽きたか、消されたかで内容を更新
		HashSet<Building> pastBurnings = new HashSet<Building>(this.burningBuildings);
		for(Building bld : pastBurnings){
			// 燃え尽きた建物更新
			if(bld.isFierynessDefined()
					&& bld.getFierynessEnum().equals(Fieryness.BURNT_OUT)){
				this.burntOutBuildings.add(bld);
				this.burningBuildings.remove(bld);
			}
			// 消火した建物更新
			else if(!bld.isOnFire()){
				this.burningBuildings.remove(bld);
				this.nears.add(bld);
			}
		}
		for(Building bld : this.buildings){
			// 所属する建物が燃えたらそちらに追加
			if(bld.isOnFire()){
				this.burningBuildings.add(bld);
				// 他のSetに所属していた場合削除
				if(this.nears.contains(bld)){
					this.nears.remove(bld);
				}
			}
			// 所属する建物で燃えてない建物以外で燃え尽きてない建物はnearsへ移行
			// おそらく燃えているところを目撃する前に消火された建物が当てはまる
			else if(bld.isFierynessDefined()
					&& !bld.getFierynessEnum().equals(Fieryness.BURNT_OUT)){
				this.nears.add(bld);
				// 他のSetに所属していた場合削除
				if(this.burningBuildings.contains(bld)){
					this.burningBuildings.remove(bld);
				}
			}
		}
	}
	
	/**
	 *  所属している建物の中で燃えている建物のHashSetを返す
	 */
	public HashSet<Building> getBurningBuildings(){
		return this.burningBuildings;
	}
	
	/**
	 * FireGroupに所属している建物を返す
	 */
	public HashSet<Building> getBuildings(){
		return this.buildings;
	}
	
	/** 所属している建物の中で燃え尽きた建物を返す */
	public HashSet<Building> getBurntOutBuildings(){
		return this.burntOutBuildings;
	}
	
	/** このグループの建物の燃えている割合を返す */
	public double getFireRate(){
		// 燃えている建物の割合を計算
		return (double)burningBuildings.size()/buildings.size();
	}
	
	/** このグループの付近で燃えそうな建物の集合を返す */
	public HashSet<Building> getNears(){
		return this.nears;
	}
	
	/** このグループの発火点となっている建物の集合を返す */
	public HashSet<Building> getIgnitions(){
		return this.ignitions;
	}
	
	/** このグループの付近で燃えそうな建物を再度セット */
	public void setNears(HashSet<Building> sets){
		this.nears.clear();
		for(Building bld : sets){
			if(!this.buildings.contains(bld)){
				this.nears.add(bld);
			}
		}
	}
	
	/** このグループの燃えている建物群を追加 */
	public void addBurningBuildings(HashSet<Building> sets){
		this.burningBuildings.addAll(sets);
		this.buildings.addAll(sets);
	}
	
	/** このグループの燃えている建物を追加 */
	public void addBurningBuilding(Building bld){
		this.burningBuildings.add(bld);
		this.buildings.add(bld);
	}
	
	/** このグループの付近で燃えそうな建物を追加 */
	public void addNear(Building bld){
		if(!this.burningBuildings.contains(bld)
			&& !this.burntOutBuildings.contains(bld)){
			this.nears.add(bld);	
		}
		this.buildings.add(bld);
	}
	
	/** このグループに所属する建物で燃えている建物があるかどうかを返す */
	public boolean isOnFire(){
		for(Building bld : this.buildings){
			if(bld.isOnFire()){
				return true;
			}
		}
		return false;
	}
	
	/** このグループ内のみでのbuildingScoreを返す */
	public double buildingScore(){
		// BuildingScoreの値を計算
		double sum = 0.0d;
		// 燃えている建物での計算
		for(Building bld : burningBuildings){
			// Fierynessの値が不明なときは無視
			if(!bld.isFierynessDefined()){
				continue;
			}
			// heating, burningのときは残量を加算
			if(bld.getFierynessEnum().equals(Fieryness.HEATING)){
				sum += 0.66d;
			}
			else if(bld.getFierynessEnum().equals(Fieryness.BURNING)){
				sum += 0.33d;
			}
			// スコア0.0はマズい気がするのでINFERNO（真っ赤）も一応加算
			else if(bld.getFierynessEnum().equals(Fieryness.INFERNO)){
				sum += 0.01d;
			}
		}
		return sum/buildings.size();
	}
	
	/**
	 * 消火済みの建物をセット。
	 * @param bld
	 */
	public void setExtinguishedBuilding(Building bld){
		this.burningBuildings.remove(bld);
		this.nears.add(bld);
	}
	
	/**
	 * このグループに所属する建物の合計面積を返す
	 * @return
	 */
	public int totalArea(){
		int total = 0;
		for(Building bld : buildings){
			// buildingの面積が不明な場合は無視
			if(!bld.isGroundAreaDefined()){
				continue;
			}
			// 各Buildingの面積を加算
			total += bld.getGroundArea();
		}
		return total;
	}
	
	/**
	 * このFireAreaの優先度を返す
	 * @return
	 */
	public int priority(){
		int priority = Integer.MAX_VALUE/2;
		// TODO Priority確認してTaskのものを利用する仕様に変更
//		int priority = Task.high;
		priority -= this.totalArea();
		priority *= this.buildingScore();
		return priority;
	}
	
	/**
	 * FireAreaを統合する
	 * @param fa
	 * @return 統合したか否か
	 */
	public boolean unionFireArea(FireArea fa){
		for(Building bld : fa.getBurningBuildings()){
			if(this.burningBuildings.contains(bld)){
				this.buildings.addAll(fa.getBuildings());
				this.burningBuildings.addAll(fa.getBurningBuildings());
				this.burntOutBuildings.addAll(fa.getBurntOutBuildings());
				this.ignitions.addAll(fa.getIgnitions());
				return true;
			}
		}
		return false;
	}
	
	public int compareTo(FireArea fa){
		return fa.priority()-this.priority();
	}
	
	/** 火災エリアが一致するかどうかをignitionを用いて判定 */
	public boolean equals(FireArea fa){
		return this.ignitions.equals(fa.getIgnitions());
	}
}
