package rione.util;

import java.util.Comparator;

import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.worldmodel.properties.IntProperty;

/**
 * IntPropertyの値でソートするためのコンパレータ<br>
 * 暗黙的にIntPropertyにキャストするので、
 * urnにIntPropertyでないPropertyを指定してはならない<br><br>
 * 使用例：<br>
 * 		List<StanderdEntity> list = new ArrayList(world.getEntitiesOfType(StanderdEntityURN.CIVILIAN)); <br>
 * 		Collections.sort(list, new IntPropertySorter(StanderdPropertyURN.HP, OrderBy.DESC));
 * <br><br>
 * この記述方法では、listに格納されたCivilianが、HPの降順にソートされる。<br>
 * ただし、Propertyが定義されていないEntityは自動的にソート後のリストの最後に配置される。
 * 
 * @author com
 *
 */
public class IntPropertySorter implements Comparator<StandardEntity> {

	/** 昇順か降順か */
	private final OrderBy order;
	
	/** ソートの基準にするPropertyのURN */
	private final StandardPropertyURN urn;
	
	/**
	 * 標準コンストラクタ
	 * @param urn ソートの基準にするPropertyのURN
	 * @param order 昇順か降順か
	 */
	public IntPropertySorter(StandardPropertyURN urn, OrderBy order) {
		this.urn = urn;
		this.order = order;
	}
	
	/**
	 * 省略コンストラクタ
	 * 自動的に昇順にソートされる
	 * @param urn ソートの基準にするPropertyのURN
	 */
	public IntPropertySorter(StandardPropertyURN urn) {
		this(urn, OrderBy.ASC);
	}
	
	@Override
	public int compare(StandardEntity a, StandardEntity b) {
		IntProperty property_a = (IntProperty)a.getProperty(urn.toString());
		IntProperty property_b = (IntProperty)b.getProperty(urn.toString());
		if (!property_a.isDefined()) {
			if (!property_b.isDefined()) {
				return 0;
			} else {
				return 1;
			}
		} else {
			if (!property_b.isDefined()) {
				return -1;
			}
		}
		return (property_a.getValue() - property_b.getValue()) * order.by();
	}

}
