package rione.viewer.component.controller;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import javax.script.Invocable;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import javax.swing.JFileChooser;

import rescuecore2.worldmodel.EntityID;
import rione.viewer.component.AdvancedViewComponent;

/*
 * JavaScript(*.js)の書き方
 * 	メソッドが毎サイクル呼ばれるのでそれを定義してやればよい
 * 	$HOME/resq.jsがあればそれを実行する
 * 	環境変数HOMEを設定すればWindowsでもおｋ
 * 	なければダイアログを出す
 * 	引数は全てサイクル数
 * 	定義しない場合は無視される
 * 	()内は値か戻り値の型
 * 
 * 	変数
 * 		debug
 * 			エラー表示をするか(true, false)
 * 	メソッド
 * 		focus(t)
 * 			フォーカスするID(整数)
 * 		visible(t)
 * 			視界内表示をするか(true, false)
 * 		plot(t)
 * 			worldmodelの座標表示
 * 		extension(t)
 * 			自由装飾を表示するか(true, false)
 * 		render(t)
 * 			自由描画を表示するか(true, false)
 * 		command(t)
 * 			コマンドを表示するか(true, false)
 * 		saveFolder(t)
 * 			保存先フォルダ(文字列)
 * 		saveImg(t)
 * 			画像を保存するか(true, false)
 * 		saveLog(t)
 * 			独自ログを出すか(true, false)
 * 		followFocus(t)
 * 			カメラ追従(true, false)
 * 		cvColor(t)
 * 			CVの色(16進)
 * 		atColor(t)
 * 			ATの色(16進)
 * 		fbColor(t)
 * 			FBの色(16進)
 * 		pfColor(t)
 * 			PFの色(16進)
 * 		update(t)
 * 			毎サイクル呼ばれる
 * 			文字列
 * 			"ファイルパス"
 * 				新しいスクリプトを起動
 * 			"newPanel"
 * 				PanelControllerを起動
 */

/**
 * スクリプトで操作するコントローラ
 * 
 * @author utisam
 * 
 */
public class ScriptController implements Controller {

    /** コントロール対象 */
    private final AdvancedViewComponent viewer;

    private boolean scriptDebug = false;
    private EntityID currentFocus = null;
    private boolean visibleEntity = false;
    private boolean customExtension = false;
    private boolean customRender = false;
    private boolean lastCommand = false;
    private boolean plotLocation = false;
    private String saveFolderName = null;
    private boolean saveImage = false;
    private boolean saveLog = false;
    private boolean followFocus = false;
    private Color cvColor;
    private Color atColor;
    private Color fbColor;
    private Color pfColor;
    private int frameWidth = 0, frameHigh = 0;

    /** スクリプト呼び出しのインタフェース */
    protected Invocable inv = null;

    /**
     * Controllerを引き継いでコンストラクト
     * 
     * @param c
     */
    public ScriptController(Controller c) {
	this(c.getViewComponent(), null);
    }

    /**
     * ControllerとScriptを引き継いでコンストラクト
     * 
     * @param c
     * @param newScript
     */
    public ScriptController(Controller c, File newScript) {
	this(c.getViewComponent(), newScript);
	setFocus(c.getFocus());
	visibleEntity = c.visibleEntity();
	customExtension = c.costomExtention();
	customRender = c.customRender();
	lastCommand = c.lastCommand();
	plotLocation = c.plotLocation();
	saveFolderName = c.getSaveFolderName();
	saveImage = c.saveImage();
	saveLog = c.saveLog();
	cvColor = c.getCivilianColor();
	atColor = c.getAmbulanceTeamColor();
	fbColor = c.getFireBrigadeColor();
	pfColor = c.getPoliceForceColor();
    }

    /**
     * 初回用コンストラクタ
     */
    public ScriptController(AdvancedViewComponent v, File newScript) {
	viewer = v;

	// スクリプトエンジン生成
	ScriptEngineManager factory = new ScriptEngineManager();
	ScriptEngine engine = factory.getEngineByName("JavaScript");

	// 評価
	try {
	    engine.eval(new FileReader((newScript == null) ? createScriptFile() : newScript));
	}
	catch (FileNotFoundException e) {
	    e.printStackTrace();
	}
	catch (ScriptException e) {
	    e.printStackTrace();
	}
	catch (NullPointerException e) {
	    try {
		engine.eval("function update(t) {return 'newPanel';}");
	    }
	    catch (ScriptException e1) {
		e1.printStackTrace();
	    }
	}

	// デバッグするかを設定
	try {
	    scriptDebug = (Boolean) engine.get("debug");
	}
	catch (ClassCastException e) {
	    System.err.println("スクリプトのdebugが不正です");
	    scriptDebug = false;
	}
	catch (NullPointerException e) {
	    scriptDebug = false;
	}

	// フレームの大きさ
	try {
	    frameWidth = ((Double) engine.get("width")).intValue();
	}
	catch (ClassCastException e) {
	    System.err.println("スクリプトのwidthが不正です");
	    frameWidth = 0;
	}
	catch (NullPointerException e) {
	    frameWidth = 0;
	}
	try {
	    frameHigh = ((Double) engine.get("high")).intValue();
	}
	catch (ClassCastException e) {
	    System.err.println("スクリプトのhighが不正です");
	    frameHigh = 0;
	}
	catch (NullPointerException e) {
	    frameHigh = 0;
	}

	// アクセスのためのインタフェースだけを保持
	inv = (Invocable) engine;
    }

    /**
     * スクリプトファイルを生成する<br>
     * "${HOME}/resq.js"を参照する<br>
     * Windowsでも環境変数設定すればおｋ
     * 
     * @return
     */
    private static File createScriptFile() {
	// resq.jsを読む
	File defaultScript = null;
	try {
	    defaultScript = new File(System.getenv("HOME"), "resq.js");
	}
	catch (NullPointerException e) {
	    defaultScript = null;
	}
	catch (SecurityException e) {
	    System.err.println("createScriptFile: permission denied");
	    defaultScript = null;
	}
	// resq.jsがあればそれを返す
	if (defaultScript != null && defaultScript.exists()) {
	    return defaultScript;
	}
	else {
	    // 無ければファイル選択ダイアログを出す
	    JFileChooser fc = new JFileChooser();
	    fc.showDialog(new Frame("*.js"), "読み込み");
	    return fc.getSelectedFile();
	}
    }

    @Override
    public Controller update(int time) {
	currentFocus = currentFocusResult(time);
	visibleEntity = visibleEntityResult(time);
	customExtension = customExtensionResult(time);
	customRender = customRenderResult(time);
	lastCommand = lastCommandResult(time);
	plotLocation = plotLocationResult(time);
	saveFolderName = saveFolderNameResult(time);
	saveImage = saveImageResult(time);
	saveLog = saveLogResult(time);
	followFocus = followFocusResult(time);
	cvColor = agentColorResult(time, "cv");
	atColor = agentColorResult(time, "at");
	fbColor = agentColorResult(time, "fb");
	pfColor = agentColorResult(time, "pf");
	return updateResult(time);
    }

    private boolean plotLocationResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("plot", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no plot method");
	}
	return result;
    }

    /**
     * スクリプトからFocusを取得
     * 
     * @param t
     * @return
     */
    private EntityID currentFocusResult(int t) {
	Integer result;
	try {
	    // Integerに直接キャストするとClassCastException
	    result = ((Double) inv.invokeFunction("focus", t)).intValue();
	    if (currentFocus == null || result != currentFocus.getValue()) {
		return new EntityID(result);
	    }
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Integer以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no focus method");
	}

	return currentFocus;
    }

    /**
     * スクリプトからvisibleEntityを取得
     * 
     * @param t
     * @return
     */
    private boolean visibleEntityResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("visible", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no visible method");
	}
	return result;
    }

    /**
     * 
     * @param t
     * @return
     */
    private boolean customExtensionResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("extension", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no extension method");
	}
	return result;
    }

    /**
     * 
     * @param t
     * @return
     */
    private boolean customRenderResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("render", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no render method");
	}
	return result;
    }

    /**
     * 
     * @param t
     * @return
     */
    private boolean lastCommandResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("command", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no command method");
	}
	return result;
    }

    /**
     * 
     * @param t
     * @return
     */
    private String saveFolderNameResult(int t) {
	Object result = null;
	try {
	    result = inv.invokeFunction("saveFolder", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no saveFolder method");
	}
	if (result == null) {
	    return null;
	}
	return result.toString();
    }

    /**
     * 
     * @param t
     * @return
     */
    private boolean saveImageResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("saveImg", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no saveImg method");
	}
	return result;
    }

    /**
     * 
     * @param t
     * @return
     */
    private boolean saveLogResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("saveLog", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no saveLog method");
	}
	return result;
    }

    public boolean followFocusResult(int t) {
	Boolean result = false;
	try {
	    result = (Boolean) inv.invokeFunction("followFocus", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Boolean以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no followFocus method");
	}
	return result;
    }

    /**
     * 
     * @param t
     * @param agent
     * @return
     */
    private Color agentColorResult(int t, String agent) {
	Integer result = null;
	try {
	    result = ((Double) inv.invokeFunction(agent + "Color", t)).intValue();
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    e.printStackTrace();
	}
	catch (ClassCastException e) {
	    // Integer以外の値
	    if (scriptDebug)
		e.printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no " + agent + "Color method");
	}
	if (result == null)
	    return null;
	return new Color(result);
    }

    /**
     * 
     * @param t
     * @return
     */
    private Controller updateResult(int t) {
	Object result = null;
	try {
	    result = inv.invokeFunction("update", t);
	}
	catch (ScriptException e) {
	    // スクリプト内エラー
	    if (scriptDebug)
		e.printStackTrace();
	    return new PanelController(this);
	}
	catch (NoSuchMethodException e) {
	    // メソッドが見つからない
	    if (scriptDebug)
		System.out.println("no update method");
	}
	if (result == null) {
	    return this;
	}
	if (result.equals("newPanel")) {
	    // PanelController生成・データを引き継ぎ
	    return new PanelController(this);
	}
	else if (result instanceof String) {
	    File newScript = new File((String) result);
	    if (newScript.exists()) {
		// ScriptController生成・データを引き継ぎ
		return new ScriptController(this, newScript);
	    }
	}
	return this;
    }

    @Override
    public EntityID getFocus() {
	return currentFocus;
    }

    @Override
    public void setFocus(EntityID id) {
	currentFocus = id;
    }

    @Override
    public boolean visibleEntity() {
	return visibleEntity;
    }

    @Override
    public boolean costomExtention() {
	return customExtension;
    }

    @Override
    public boolean customRender() {
	return customRender;
    }

    @Override
    public boolean lastCommand() {
	return lastCommand;
    }

    @Override
    public String getSaveFolderName() {
	if (saveFolderName == null || saveFolderName.equals("")) {
	    return null;
	}
	return saveFolderName;
    }

    @Override
    public boolean followFocus() {
	return followFocus;
    }

    @Override
    public Color getCivilianColor() {
	return cvColor;
    }

    @Override
    public Color getAmbulanceTeamColor() {
	return atColor;
    }

    @Override
    public Color getFireBrigadeColor() {
	return fbColor;
    }

    @Override
    public Color getPoliceForceColor() {
	return pfColor;
    }

    @Override
    public boolean saveImage() {
	return saveImage;
    }

    @Override
    public boolean saveLog() {
	return saveLog;
    }

    @Override
    public AdvancedViewComponent getViewComponent() {
	return viewer;
    }

    @Override
    public Dimension getFrameSize() {
	if (frameWidth == 0 || frameHigh == 0) {
	    return null;
	}
	return new Dimension(frameWidth, frameHigh);
    }

    @Override
    public boolean plotLocation() {
	return plotLocation;
    }

}
